#!/usr/bin/perl -w
use strict;    
use warnings;  
use utf8;

use POSIX qw(locale_h);
my $old_locale = setlocale(LC_CTYPE);

my $inputFile = "Lexique380_7cols.txt";
my $outputFile = "Lexique380_7cols_improved.txt";
my $rejectsFile = "Lexique380_7cols_rejects.txt";


open (INPUT,  "<:utf8" , $inputFile)  or die "Can't read inputFile $!";
open (OUTPUT, ">:utf8" , $outputFile) or die "Can't write outputFile  $!";
open (REJECTS, ">:utf8" , $rejectsFile) or die "Can't write outputFile  $!";

# print OUTPUT localtime . "\n\n";

while(<INPUT>) {

	if (1) { # rand(1)<.04
		chomp;
		my @temp = split(/\t/);
	
		my $spelling      = $temp[0];
		my $transcription = $temp[1];
		my $lemma         = $temp[2];
		my $POS           = $temp[3] || "";
		
		my $ipa_transcription = toIPA($transcription);

		# $spelling = improveSpelling($spelling, ($temp[6]||""));
		my $IPAified_spelling = IPAifiedSpelling($spelling, ($temp[6]||""));

		my @spell = split(//,$IPAified_spelling);
		my @trans = split(/ /,$ipa_transcription);

		my ($al1, $al2) = Needle(\@spell, \@trans);
		my $output = printAlign($al1, $al2);
		
		# improve bad alignments
		$output =~ s/ *$//g;
		if ($output =~ /. ø$/ and $spelling =~ /e$/) {
			print $spelling . "\n";
			$output =~ s/(.) ø/$1 ə/g;
		}
		$output =~ s/\(s\) \(ə\) s$/s (ə) (s)/g;
		$output =~ s/\(z\) \(ə\) s$/s (ə) (s)/g;
		$output =~ s/\(t\) \(ə\) t$/t (ə) (t)/g;
		$output =~ s/\(ə\) \(s\) ɛ s$/ɛ s (ə) (s)/g;
		$output =~ s/\(s\) \(ə\) s ɛ s$/s ɛ s (ə) (s)/g;
		
		
		# print OUTPUT $_ . "\t" . printAlign($al1, $al2) . "\n" if (($temp[6]||"")=~/3p/);
		if ($output =~ /\[/ and $output !~ /\[i,j\]/ and $output !~ /\[y,ɥ\]/ and $output !~ /\[u,w\]/) {  
			print REJECTS $spelling. "\t" . $IPAified_spelling .  "\t" .  $ipa_transcription . "\t" . $output . "\n" ; 
			print OUTPUT $spelling. "\t" . $output . "\t" . POS($POS) . "\n" ; 
		} else {
			$output = clearConsonants($output);
			#if ($output =~ /\([^əstʁzdkpbŋlf]\)/) {
				print OUTPUT $spelling. "\t" . $output . "\t" . POS($POS)  . "\n" ; 
			#}
		}
	}

}
close(INPUT) or die "OMG Can't close stimuli file $!";
close(OUTPUT) or die "OMG Can't close output file $!";
close(REJECTS) or die "OMG Can't close output file $!";

sub clearConsonants {
	my $s = shift(@_);
	while ($s =~ /\([pbfvmtdsznʃʒɲkɡŋxhlʁ]\) [\(]?[pbfvmtdsznʃʒɲkɡŋxhlʁ]/) {
		$s =~ s/\([pbfvmtdsznʃʒɲkɡŋxhlʁ]\) ([\(]?[pbfvmtdsznʃʒɲkɡŋxhlʁ])/$1/g;
	}
	return $s;
}

sub POS {

	my $s = shift(@_);

	$s =~ s/^(VER|AUX|NOM|ADJ|ADJ:ind|ADJ:num|ADV|ONO)$/LEX/;
	$s =~ s/^(ADJ:dem|ADJ:int|ADJ:pos|ART:def|ART:ind|CON|LIA|PRE|PRO:dem|PRO:ind|PRO:int|PRO:per|PRO:pos|PRO:rel)$/FNC/;

	# if($s !~ /^(LEX|FNC)$/) { print $s . "\n";}
	return $s;
}

sub toIPA {
	my $s = shift(@_);

#	$s =~ s/-/- /g; # add space after a dash
#	$s =~ s/'/' /g; # add space after apostrophe
	
	$s =~ s/(.)/$1 /g; # add spaces between characters
	
	$s =~ s/§/õ/g;
	$s =~ s/1/ũ/g;
	$s =~ s/5/ẽ/g;
	$s =~ s/@/ã/g;
	$s =~ s/2/ø/g;
	$s =~ s/9/œ/g;
	$s =~ s/E/ɛ/g;
	$s =~ s/O/ɔ/g;
	$s =~ s/°/ə/g;
	$s =~ s/8/ɥ/g;
	$s =~ s/3/ə/g;

	$s =~ s/G/ŋ/g;
	$s =~ s/N/ɲ/g;
	$s =~ s/Z/ʒ/g;
	$s =~ s/S/ʃ/g;
	$s =~ s/R/ʁ/g;
	$s =~ s/g/ɡ/g;

	return $s;

}

sub IPAifiedSpelling {
	my $s = shift(@_);
	my $verb_info = shift(@_);

	$s = uc($s);

	if (($verb_info||"")=~/3p/) {
		$s =~ s/ENT$/ət/g;
		$s =~ s/ONT$/õt/g;
	}
 	$s =~ s/CCH/k/g;
 	$s =~ s/CH([LRMN])/k$1/g;
 	$s =~ s/S?CH/ʃ/g;
 	$s =~ s/PH/f/g;
 	$s =~ s/X([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])/ɡz$1/g;
 	$s =~ s/XC([EÉÈÊËIÍÌÎÏə])/ks$1/g;
 	$s =~ s/X$/s/g;
 	$s =~ s/X/ks/g;
 	$s =~ s/S?C([IÍÌÎÏEÉÈÊËYə])/s$1/g;
 	$s =~ s/CC?/k/g;
 	$s =~ s/GN/ɲ/g;
 	$s =~ s/NG$/ŋ/g;
	$s =~ s/[UO]EILL?/œj/g;
	$s =~ s/EUILL?/œj/g; 
 	$s =~ s/UË/y/g;
 	$s =~ s/GE([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜœ])/ʒ$1/g;
 	$s =~ s/G([IÍÌÎÏEÉÈÊËYə])/ʒ$1/g;
 	$s =~ s/GU/ɡ/g;
 	$s =~ s/GG?/ɡ/g;
 	$s =~ s/JJ?/ʒ/g;
 	$s =~ s/QU/k/g;
 	$s =~ s/Q/k/g;
 	$s =~ s/([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])S([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])/$1z$2/g;
 	$s =~ s/SS?/s/g;
 	$s =~ s/Ç/s/g;
 	$s =~ s/TIO/sIO/g;

 	$s =~ s/([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])ILL([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])/$1j$2/g;
 	$s =~ s/ILL([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])/ij$1/g;

 	$s =~ s/OI[MN]([^MNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/wẽ$1/g;
 	$s =~ s/[AE]I[MN]([^MNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/ẽ$1/g;
 	$s =~ s/EU[MN]([^MNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/ũ$1/g;
 	$s =~ s/OU[MN]([^MNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/un$1/g;

 	$s =~ s/[AÁÀÂÄ][MN]([^wMNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/ã$1/g;
 	$s =~ s/[OÓÒÔÖ][MN]([^wMNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/õ$1/g;
 	$s =~ s/[UÚÙÛÜ][MN]([^wMNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/ũ$1/g;
 	$s =~ s/[IÍÌÎÏY][MN]([^wMNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/ẽ$1/g;
 	$s =~ s/I[EÉÈÊË][MN]([^wMNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/iẽ$1/g;
 	$s =~ s/[EÉÈÊË][MN]([^wMNAÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi]|$)/ã$1/g;

 	$s =~ s/([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi])Y([AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜYəõãũẽi])/$1Ij$2/g;

 	$s =~ s/[AO]W/o/g;
 	$s =~ s/W/w/g;

 	$s =~ s/OEU/œ/g;
 	$s =~ s/EAU/o/g;
 	$s =~ s/OO/o/g;
 	$s =~ s/EE/i/g;
 	$s =~ s/EA/i/g;
 	$s =~ s/AE/e/g;
 	$s =~ s/O[IÎ]/wa/g;
 	$s =~ s/A[IÎ]/e/g;
 	$s =~ s/AÏ/aj/g;
 	$s =~ s/A[UÛ]/o/g;
 	$s =~ s/(^|[^O])U[IÍÌÎÏ]/$1ɥi/g;
 	$s =~ s/È/ɛ/g;
 	$s =~ s/EI/ɛ/g;
 	$s =~ s/[ÉÊË]/e/g;
 	$s =~ s/Ô/o/g;
 	$s =~ s/E[UÚÙÛÜ]/ø/g;
 	$s =~ s/øR/œʁ/g;

# 	$s =~ s/OU(ə|õ|ũ|ẽ|ã|a|ɛ|e|i|œ|ø|y|ɔ|o|u|[AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])/w$1/g; # don't create glides
 	$s =~ s/(^|[^OÓÒÔÖ])[UÚÙÛÜ]/$1y/g;
 	$s =~ s/O[UÚÙÛÜ]/u/g;

 	$s =~ s/[AE]Y$/ɛ/g;
 	$s =~ s/Y$/i/g;
 	$s =~ s/Y/i/g;


 	$s =~ s/[AÁÀÂÄ]/a/g;
 	$s =~ s/[IÍÌÎÏ]/i/g;




	$s =~ s/Ñ/ɲ/g;
	$s =~ s/BB?/b/g;
	$s =~ s/DD?/d/g;
	$s =~ s/FF?/f/g;
	$s =~ s/KK?/k/g;
	$s =~ s/LL?/l/g;
	$s =~ s/MM?/m/g;
	$s =~ s/NN?/n/g;
	$s =~ s/PP?/p/g;
	$s =~ s/RR?/ʁ/g;
	$s =~ s/TT?/t/g;
	$s =~ s/VV?/v/g;
	$s =~ s/ZZ?/z/g;
 	$s =~ s/H//g;
 	$s =~ s/ //g;
 	$s =~ s/-//g;
 	$s =~ s/'//g;
	return $s;
}



sub printAlign {

	my @seq1 = @{shift(@_)};
	my @seq2 = @{shift(@_)};

	if (scalar(@seq1) != scalar(@seq2)) {return "can't print arrays of different lengths.";}

	my $str;
	for (my $i=0; $i<@seq1; $i++) {
		if ($seq1[$i] eq $seq2[$i]) {
			$str .=    $seq1[$i] . " ";
		} else {
			my $temp = $seq1[$i].$seq2[$i]; # print $temp . "\n";
			if ($temp =~ /(õə|tz|eø|Oœ|kʁ|øɔ|øu|ɲɡ|yɔ|eẽ|ũø|ãe|dt|œy|ks|nɲ|ʒx|õø|iẽ|ãə|ũa|ao|ũy|ji|ie|ei|ʒʃ|ʃʒ|ʒʁ|aẽ|ẽa|iœ|œi|aɔ|ɔa|ũœ|œũ|wv|vw|Ou|ia|ai|ɡj|jɡ|ʒj|jʒ|Oa|ou|Oõ|oa|ũõ|kt|ẽi|ãɛ|Ei|Eɛ|Ee|Eə|Eø|Eœ|eɛ|ɛe|ɛẽ|Oo|Oɔ|oɔ|ɔo|øœ|œø|ʃk|kʃ|ẽã|ãẽ|ẽɛ|sʃ|ʃs|bp|pb|zs|sz|ts|st|ɡk|kɡ|jl|lj|yœ|Ea|Eã|yø|øy|yu|ɲn|ŋɡ|ɡŋ|ʒɡ|ɡʒ|õã|ãõ|aã|ea|ãa|ũɔ|ũu|õo|õɔ|ae|aɛ|iɛ|eə|ɛa)/) {
				$str .=    $seq2[$i] . " ";
			} else {
				if ($seq2[$i] eq "•") {
					if ($seq1[$i] eq "E") {
						$str .= "(" . "ə" . ") "; # extra from spelling inserted
					} else {
						$str .= "(" . $seq1[$i] . ") "; # extra from spelling inserted
					}
				} else {
					if ($seq1[$i] eq "•") {
						$str .= "«" . $seq2[$i] . "» "; # extra from IPA inserted
					} else {
						$str .= "[" . $seq1[$i] . "," . $seq2[$i] . "] ";
					}
				}
			}
		}
	}
	return $str;

}



sub Needle {
	
	my @seq1 = @{shift(@_)};
	my @seq2 = @{shift(@_)};

	# scoring scheme (too complicated to figure out by now)
	my $CMATCH    =  4;  # 
	my $VMATCH    =  3;  # 
	my $PERFECT   =  4;  # 
	my $CCMISMATCH = -3; # 
	my $VVMISMATCH = -3; # 
	my $CVMISMATCH = -8; # 
	my $GAP      = -3;   # 
	
	# initialization
	my @matrix;
	$matrix[0][0]{score}   = 0;
	$matrix[0][0]{pointer} = "none";
	for(my $j = 1; $j <= @seq1; $j++) {
		$matrix[0][$j]{score}   = $GAP * $j;
		$matrix[0][$j]{pointer} = "left";
	}
	for (my $i = 1; $i <= @seq2; $i++) {
		$matrix[$i][0]{score}   = $GAP * $i;
		$matrix[$i][0]{pointer} = "up";
	}
	
	# fill
	for(my $i = 1; $i <= @seq2; $i++) {
		for(my $j = 1; $j <= @seq1; $j++) {
			my ($diagonal_score, $left_score, $up_score);
	
			# calculate match score
			my $letter1 = $seq1[$j-1];
			my $letter2 = $seq2[$i-1];   
			my $temp = $letter1 . $letter2;                          
			if ($letter1 eq $letter2) {
				if ($letter1 =~ /ə|õ|ũ|ẽ|ã|a|ɛ|e|i|œ|ø|y|ɔ|o|u|[AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ]/) {
					$diagonal_score = $matrix[$i-1][$j-1]{score} + $VMATCH + $PERFECT;
				} else {
					$diagonal_score = $matrix[$i-1][$j-1]{score} + $CMATCH + $PERFECT;
				}
			}
			else {
				if ($temp =~ /(ə|õ|ũ|ẽ|ã|a|ɛ|e|i|œ|ø|y|ɔ|o|u|[AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])(ə|õ|ũ|ẽ|ã|a|ɛ|e|i|œ|ø|y|ɔ|o|u|[AÁÀÂÄEÉÈÊËIÍÌÎÏOÓÒÔÖUÚÙÛÜ])/) {
					if ($temp =~ /ɛe|eɛ|ɔO|Oɔ|oɔ|ɔo|oO|Oo|yø|øy|øœ|œø|œy|yœ|ɛẽ|ẽɛ/) {
						$diagonal_score = $matrix[$i-1][$j-1]{score} + $VMATCH + $PERFECT;
					} else {
						if ($temp =~ /əɛ|əe/) {
							$diagonal_score = $matrix[$i-1][$j-1]{score} + $VVMISMATCH;
						} else {
							$diagonal_score = $matrix[$i-1][$j-1]{score} + $VMATCH;
						}
					}
				} else {
					if ($temp =~ /(pb|bp|sz|zs|kɡ|ɡk|td|dt|fv|vf|ʃʒ|ʒʃ|lj|jl|ij|ji|uw|wu|ɥy|yɥ|ʒɡ|ɡʒ|ɲɡ)/) {
						$diagonal_score = $matrix[$i-1][$j-1]{score} + $CMATCH;
					} else {
						if ($temp =~ /[pbfvmtdsznʃʒɲkɡŋxhlʁjɥw][pbfvmtdsznʃʒɲkɡŋxhlʁjɥw]/) {
							if ($temp =~ /sn|sm|sŋ|sk|sj/) {
								$diagonal_score = $matrix[$i-1][$j-1]{score} + $CVMISMATCH;
							} else {
								$diagonal_score = $matrix[$i-1][$j-1]{score} + $CCMISMATCH;
							}
						} else {
							$diagonal_score = $matrix[$i-1][$j-1]{score} + $CVMISMATCH;
						}
					}
				}
			}
	
	
			# calculate gap scores
			$up_score   = $matrix[$i-1][$j]{score} + $GAP;
			$left_score = $matrix[$i][$j-1]{score} + $GAP;
	
			# choose best score
			if ($diagonal_score >= $up_score) {
				if ($diagonal_score >= $left_score) {
					$matrix[$i][$j]{score}   = $diagonal_score;
					$matrix[$i][$j]{pointer} = "diagonal";
				}
			else {
					$matrix[$i][$j]{score}   = $left_score;
					$matrix[$i][$j]{pointer} = "left";
				}
			} else {
				if ($up_score >= $left_score) {
					$matrix[$i][$j]{score}   = $up_score;
					$matrix[$i][$j]{pointer} = "up";
				}
				else {
					$matrix[$i][$j]{score}   = $left_score;
					$matrix[$i][$j]{pointer} = "left";
				}
			}
		}
	}
	
	# trace-back
	
	my @align1;
	my @align2;
	
	# start at last cell of matrix
	my $j = scalar(@seq1);
	my $i = scalar(@seq2);
	
	while (1) {
		last if $matrix[$i][$j]{pointer} eq "none"; # ends at first cell of matrix
	
		if ($matrix[$i][$j]{pointer} eq "diagonal") {
			unshift(@align1, $seq1[$j-1]);
			unshift(@align2, $seq2[$i-1]);
			$i--;
			$j--;
		}
		elsif ($matrix[$i][$j]{pointer} eq "left") {
			unshift(@align1, $seq1[$j-1]);
			unshift(@align2, "•");
			$j--;
		}
		elsif ($matrix[$i][$j]{pointer} eq "up") {
			unshift(@align1, "•");
			unshift(@align2, $seq2[$i-1]);
			$i--;
		}    
	}
	
	return(\@align1, \@align2);
}

