﻿#!/usr/bin/perl -w

use strict;    # use this for every damn script you ever write
use warnings;  # use this for every damn script you ever write

use Fcntl qw(:flock); # lets you lock files to prevent others from using them while they're open
use utf8;


# input files
my $corpusFile  = "verlaine.txt" ;
#my $corpusFile  = "baudelaire.txt" ;
my $lexiconFile = "../lexique/Lexique380_7cols_improved.txt" ;
my $lexiconFile2 = "../lexique/added_words.txt" ;

# output files
my $outputFileAlex   = ($corpusFile =~ /(.+)\.txt$/)[0] . "_tagged_alex.txt";
my $outputFileOthers = ($corpusFile =~ /(.+)\.txt$/)[0] . "_tagged_others.txt";
my $notfoundFile = ($corpusFile =~ /(.+)\.txt$/)[0] . "_notfound.txt";
my %notfound;
#my $ambigFile = ($corpusFile =~ /(.+)\.txt$/)[0] . "_ambig.txt";
#my %ambig;
my @stats;
#$stats[0] = "sonority\ttype\tword\tmeter\toriginal";
my $statsFile = ($corpusFile =~ /(.+)\.txt$/)[0] . "_stats.txt";

my $vowel = "[õũẽãəaɛeiɔouœøy]";
my $consonant = "[pbfvmtdsznʃʒɲkɡŋxhlʁ<]";
my $small = " ▷ ";
my $big   = " ▶ ";

#my $small = " x";
#my $big   = " x";

my %lexicon;
my %lexiconPOS;
for my $counter (1..2) {
	
	my $file = ($counter==1) ? $lexiconFile : $lexiconFile2 ;

	open (INPUT,  "<:utf8" , $file)       or die "Can't read dictionary file $!";
	flock(INPUT, LOCK_SH) or die "$!";  # lock file for shared use
	
	while (<INPUT>) {
		chomp;
		my @temp = split(/\t/);

		my $spelling = $temp[0];
		my $transcription = $temp[1];
		my $POS = $temp[2] || "";

		if (exists($lexiconPOS{$spelling})) {
			$lexiconPOS{$spelling} = ($lexiconPOS{$spelling} eq "FNC") ? "FNC" : $POS;  
		} else {
			$lexiconPOS{$spelling} = $POS;  
		}

		$transcription =~ s/«//g;
		$transcription =~ s/»//g;
		$transcription =~ s/ɛ \(ə\)/ɛ/g;

		my @transcriptions;
		push (@transcriptions, $transcription);
		while ($transcriptions[0] =~ /^(.*)\[(.)\,(.)\](.*)$/) {
			my @newtranscriptions;
			for my $t (@transcriptions) {
				$t =~ /^(.*)\[(.)\,(.)\](.*)$/;
				push (@newtranscriptions, $1 . $2 . $4);
				push (@newtranscriptions, $1 . $3 . $4);
			}
			@transcriptions = @newtranscriptions;
		}
		
		if (!exists($lexicon{$spelling})) {
			$lexicon{$spelling} = [];
		}
		for my $t (@transcriptions) {
			unless($t ~~ @{ $lexicon{$spelling} }) {
				push( @{ $lexicon{$spelling} } , $t );
			} 
		}

	}
	close(INPUT)  or die "$!";
}







open (INPUT,   "<:utf8" , $corpusFile)       or die "Can't read input file $!";
open (OUTPUT,  ">:utf8" , $outputFileAlex)   or die "Can't write to output file $!";
open (OUTPUT2, ">:utf8" , $outputFileOthers) or die "Can't write to output file $!";

flock(INPUT,  LOCK_EX) or die "$!"; # lock file for exclusive use
flock(OUTPUT, LOCK_EX) or die "$!"; # lock file for exclusive use
flock(OUTPUT2,LOCK_EX) or die "$!"; # lock file for exclusive use

my $skipped = 0;
my @stanzaTranscription;
my @stanzaOriginal;
my @lineCount;

while (<INPUT>) {

	chomp; # remove the "end of line" character from the current word

	# lines to skip: those that
	# 			start with # 
	# 			have a digit
	# 			empty / spaces only
	# 			starts with 12 or more spaces
	# 			have two or more capital letters in a row
	# 			end in space + I or V (roman numberals)
	# 			have . . . 
	if (/^#|\d|^ *$|^ {12,}|[A-ZÉ]{2,}| [IV]$|\. \. \. /) {
		# skipped line
		# print OUTPUT "# " . $_ . "\n";
		$skipped = 1;
	} else {
		# accumulate material for the current stanza
		push(@stanzaOriginal, $_);
		my @candidates = @{ makeCandidates($_) } ;
		push(@stanzaTranscription, \@candidates);
		$skipped = 0;
	}

	if ($skipped and @stanzaTranscription) {

		# there is a stanza to print, let's do so 
		if (scalar(@stanzaOriginal) != scalar(@stanzaTranscription)) {
			# this should never happen
			# just checking that there is the same number of lines in the original and transcription
			print "@stanzaOriginal\n\n\n";
		}
	
		# collect info about the number of syllables in the line of the current stanza
		my @linelengths;
		for my $line (@stanzaTranscription) {
			my @l = ();
			for my $cand (@{$line}) { push (@l, vowelCount($cand)); }
			push (@linelengths , \@{ unique(\@l) } );
		}
		my @intersect = @{ multiIntersect(\@linelengths) };
		
		if (join("", @intersect) eq "12" or join("", @intersect) eq "1312" or (scalar(@stanzaOriginal)==1 and "12"~~@intersect)) {

			# process alexandrine

			push(@lineCount, scalar(@stanzaOriginal)); # for displaying the line numbers

			# print each line to the readable output and to the R file
			for (my $line_in_stanza=0; $line_in_stanza<@stanzaTranscription; $line_in_stanza++) {

				# print original line
				print OUTPUT $stanzaOriginal[$line_in_stanza] ."\n"; 

				my $line = $stanzaTranscription[$line_in_stanza];
				
				$line = removeNon12VowelCandidates($line);	 # remove candidates that don't have 12 vowels, 
				$line = keepShortestCandidate($line);        

				# at this point there should only be one candidate here, 
				# but we still want the code to work if there is more than one
				for my $cand (@{$line}) {
					if (vowelCount($cand)==12) {
						print OUTPUT $cand . "\n"; # transcribed line
					}
				}
				print OUTPUT "\n";
				
				# make R-readable output
				my @candidates = @{ processAlex($line, scalar(@lineCount), $line_in_stanza+1) };
				if (scalar(@candidates)==1) {
					# everything is dandy, to be printed later
					push(@stats, $candidates[0]);
				} else {
					# this should never happen - we don't want any given line to be counted in the stats more than once
					print @candidates;
				}
			}
			print OUTPUT "stanza number " . scalar(@lineCount) . ", " ;
			print OUTPUT scalar(@stanzaOriginal) . " line" . ((scalar(@stanzaOriginal)>1)?"s":"") . "\n\n";
			print OUTPUT "\n\n\n";
			
		} else {
			# print non-alexandrines
			for my $line (@stanzaOriginal) {
				print OUTPUT2 "orig:  " . $line ."\n";
			}
			print OUTPUT2 "\n";
			for my $line (@stanzaTranscription) {
				for my $cand (@{$line}) {
					print OUTPUT2 "trans(" . vowelCount($cand) . "): " . $cand . "\n";
				}
				print OUTPUT2 "\n";
			}
			print OUTPUT2 "\n";
			print OUTPUT2 "intersection: " . join(", ", @intersect) . "\n";
			print OUTPUT2 "non-alex (".scalar(@stanzaOriginal)." lines): " ;
			for my $l (@linelengths) { print OUTPUT2 join (",", @{$l}) . "  -  "; };
			print OUTPUT2 "\n\n\n";
			
		}
		@stanzaTranscription = ();
		@stanzaOriginal = ();
	}



	


}

#summary
my $total = 0;
for my $i (@lineCount) {
	$total += $i;
}
print OUTPUT join(", " , @lineCount) . "\n";
print OUTPUT "total stanzas: " . scalar(@lineCount) ."\n";
print OUTPUT "total lines: " . $total ."\n\n\n";

close(INPUT)   or die "$!";
close(OUTPUT)  or die "$!";
close(OUTPUT2) or die "$!";

if (%notfound) {
	open (OUTPUT, ">:utf8" , $notfoundFile) or die "Can't write to output file $!";
	flock(OUTPUT,LOCK_EX) or die "$!"; # lock file for exclusive use
	for (sort keys %notfound) {
		print OUTPUT "$_\t";
		s/(.)/$1 /g; # add spaces between characters
		print OUTPUT "$_\tLEX\n";
	}
	close(OUTPUT) or die "$!";
}


open (OUTPUT, ">:utf8" , $statsFile) or die "Can't write to output file $!";
flock(OUTPUT,LOCK_EX) or die "$!"; # lock file for exclusive use
print OUTPUT join("\t", "stanza", "line", "in.line",  "POS", "in.word",  "vowel", "interval", "consonants"  ) ."\n";
for my $line (@stats) {
	for my $interval (@{$line}) {
		print OUTPUT join("\t", @{$interval}) . "\n";
	}
}
close(OUTPUT) or die "$!";


######################################
### done 
######################################















sub removeNon12VowelCandidates {

	my @candidates = @{shift(@_)};
	my @newCandidates;
	for my $cand (@candidates) {
		if (vowelCount($cand)==12) {
			push(@newCandidates,  $cand);
		}
	}
	return (\@newCandidates);
}

sub keepShortestCandidate {

	# we don't want to this function to change type; it accepts an array and returns an array
	my @candidates = @{shift(@_)};
	my $shortest = $candidates[0];	
	for my $cand (@candidates) {
		if (length($cand)>length($shortest)) {
			$shortest = $cand;
		}
	}
	@candidates = ($shortest); 
	return (\@candidates);
}



sub processAlex {
	my @candidates = @{shift(@_)};
	my $stanza_number = shift(@_);
	my $line_in_stanza = shift(@_);
	my $out1 = "";
	my $out2 = "";

	my @newcandidates;
	for my $cand (@candidates) {

		my $processedLTR = "";
		# first we go forward, adding pipes for intervals,
		# and marking the position of each vowel in the line (1-13)
		# also the position in the stanza and in the number of the stanza
		my $vowel_in_line = 1;
		for (my $i=0; $i<length($cand); $i++) {
			my $char = substr($cand, $i, 1);
			if ($char =~ /$vowel/) {
				$processedLTR .= "| " . $stanza_number . " " . $line_in_stanza . " " . $vowel_in_line . " " . $char;
				$vowel_in_line++;  
			} else {
				$processedLTR .= $char;
			}
		}
		# now we go from the end backwards, adding lexical category
		# and position of vowel in the word (0 for final schwa, 1 for stress)
		my $processedRTL = "";
		my $position_in_word = 0;
		my $lexical_category;
		for (my $i=length($processedLTR)-1;  $i>=0; $i--) {
			my $char = substr($processedLTR, $i, 1);
			if ($char eq trim($big))   { $lexical_category = "lex"; $position_in_word = 0; };
			if ($char eq trim($small)) { $lexical_category = "fnc"; $position_in_word = 0; };
			if ($char =~ /$vowel/) {
				unless($char eq "ə" and $position_in_word==0) {
					$position_in_word++;
				}
				$processedRTL = $lexical_category . " " . $position_in_word . " " . $char . $processedRTL;
			} else {
				$processedRTL = $char . $processedRTL;
			}
		}
		# we have all the info, time to split into intervals and prep
		# first we remove word boundary markers
		$processedRTL = " " . $processedRTL . " ";
		$processedRTL =~ s/ +/  /g;
		$processedRTL =~ s/($small|$big)//g;
		$processedRTL =~ s/ +/ /g;
		my @intervals = split(/\|/, $processedRTL);
		# get rid of the consonants that precede the first interval
		if ($intervals[0] !~ /$vowel/) {
			shift(@intervals); 
		}
		# split intervals into useful units
		for (my $i=0; $i<@intervals; $i++) {
			my @i = split(/ /, trim($intervals[$i]));
			my @j = (shift(@i), shift(@i), shift(@i), shift(@i), shift(@i), $i[0], join("",@i));
			if(@i>1) {
				shift(@i);
				push(@j, join("",@i));
			} else {
				push(@j,"");
			}
			#print join("\t",@j) . "\t" . $cand . "\n";
			$intervals[$i] = \@j;
		}
		# print $cand . "\n" . join("___", @intervals) . "\n\n";
		push(@newcandidates, \@intervals);
	}

	
# 	my @newcandidates;
# 	for my $cand (@candidates) {
# 		$cand =~ s/$consonant|[wjɥ]//g;
# 		if (vowelCount($cand)==12) {
# 			$cand = " " . $cand . " ";
# 			$cand =~ s/ +/  /g;
# 			$cand =~ s/($small|$big)($small|$big)*/$1/g;
# 			$cand =~ s/^ *($small|$big)//g;
# 			$cand =~ s/ //g;
# 			push(@newcandidates, $cand);
# 		}
# 	}
# 	@newcandidates = @{ unique(\@newcandidates) };
# 	$out1 = "proc:  " . join("," , @newcandidates) . "\n"; 
# 	
# 
# 	for my $cand (@newcandidates) {
# 		my @c = split(//, $cand);
# 		my $category      = "";
# 		my $word_position = 0;
# 		my $line_position = 13;
# 		for (my $i=scalar(@c)-1;$i>=0;$i--){
# 			if(" ".$c[$i]." " eq $big) {
# 				$category = "lex";
# 				$word_position = 0;
# 			} 
# 			if(" ".$c[$i]." " eq $small) {
# 				$category = "fnc";
# 				$word_position = 0;
# 			}
# 			if($c[$i] =~ /$vowel/) {
# 				if ($c[$i] eq "ə" and $word_position==0) {
# 					if ($line_position<13) {
# 						$line_position--;
# 					}
# 				} else {
# 					$word_position++;
# 					$line_position--;
# 				}
# 				$c[$i] = "[" . join(",", $c[$i], $word_position, $category, $line_position, $line_in_stanza, $stanza_number) . "]";
# 			}
# 		}
# 		$cand = join(" ", @c);
# 	}
# 	for my $cand (@newcandidates) {
# 		my @words = split(/$big|$small/, $cand." ");
# 		for (my $word_number=0; $word_number<@words; $word_number++) {
# 			my $word = $words[$word_number];
# 			my @vowels = split(" ", $word);
# 			for (my $vowel_number=0; $vowel_number<@vowels; $vowel_number++) {
# 				$vowels[$vowel_number] =~ /^(\[.,)(.*\])/;
# 				if ($1 and $2) {
# 					$vowels[$vowel_number] = $1 . ($word_number+1) . "," . $2;
# 				} else {
# 					print $cand ."\t" . $word . "\t" . $vowels[$vowel_number] ."\n";
# 				}
# 			}
# 			$words[$word_number] = join(" ",@vowels);
# 		}
# 		$cand = join(" ", @words);
# 	}
# 	$out2 = "proc2: " . join(" # " , @newcandidates) . "\n"; 
# 	
	return \@newcandidates;
}





sub makeCandidates {

	my $s = shift(@_);
	
	my @candidates;
	$candidates[0] = "";

	# remove punctuation
	$s =~ s/[\!\,\.\?\:\;_\(\)\[\]]//g; 
	$s =~ s/« *//g;
	$s =~ s/ *»//g;
	$s =~ s/^(-| )+//g;
	$s =~ s/-/ /g;
	$s =~ s/ +/ /g;
	$s =~ s/'/' /g;
	
 	my @temp = split(/ /, $s);
 	for my $word (@temp) {
 		$word = lc($word);
 		if (exists($lexicon{$word})) {
 			my $POS = ($lexiconPOS{$word} eq "FNC") ? $small : $big; 
 			my @transcriptions = @{ $lexicon{$word} };
 			# duplicate candidate list if there are more than one transcriptions
 			my @newcandidates;
 			for my $trans (@transcriptions) {
 				for my $cand (@candidates) {
	 				push(@newcandidates, $cand . $trans . $POS)
 				}
 			}
 			@candidates = @newcandidates;
 		} else {
			$notfound{lc($word)} = 1;
			for my $cand (@candidates) {
				$cand .= "   (NOT FOUND: " . $word . ")   ";
			}
 		}
 	}
 	# adjust CV, remove some unneeded candidates
	my @newcandidates;
	for my $cand (@candidates) {
		$cand = adjustCV($cand);
		my $vcount = vowelCount($cand);
		if (scalar(@candidates)==1 or (1 or $vcount==12 or $vcount==11 or $vcount==10 or $vcount<9)) {
			push (@newcandidates, $cand);
			# print OUTPUT "transcription (" . $vcount . "): " . $cand . "\n";
		}
	}
	# print OUTPUT "\n";
	@candidates = @newcandidates;
	return (\@candidates);

}








sub adjustCV {

	my $s = shift(@_);

	$s =~ s/ \($consonant\)($big|$small)($consonant|$)/$1$2/g;
	$s =~ s/ \(ə\)($big|$small)($vowel|[wjɥ])/$1$2/g;
	$s =~ s/ \(($consonant)\)($big|$small)($vowel|[wjɥ])/ $1$2$3/g;
	$s =~ s/\(ə\)/ə/g;

	$s =~ s/ \($consonant\) ($consonant)/ $1/g;

	return $s;

}

sub vowelCount {

	my $s = shift(@_);

	$s =~ s/($big|$small)//g;
	$s =~ s/($consonant|[wjɥ])//g;
	$s =~ s/<//g;
	$s =~ s/ //g;

	$s =~ s/ə$//g;

	return length($s) # . $s;

}

sub intersect {
	my @a = @{shift(@_)};
	my @b = @{shift(@_)};
	my %union;
	my %isect;
	foreach my $e (@a, @b) {
		$union{$e}++ && $isect{$e}++
	}
	my @union = keys %union;
	my @isect = keys %isect;
	return \@isect;
}

sub multiIntersect {
	my @lists = @{shift(@_)};
	my @intersect;
	@intersect = @{$lists[0]};
	for (my $i=1; $i<@lists; $i++){
		@intersect = @{ intersect(\@intersect, \@{$lists[$i]}) };
	}
	return \@intersect;
}


sub unique {
	my @list = @{shift(@_)};
	my %seen = ();
	my @uniq;
	foreach my $item (@list) {
    	push(@uniq, $item) unless $seen{$item}++;
	}
	return \@uniq;
}

sub trim {
    (my $s = $_[0]) =~ s/^\s+|\s+$//g;
    return $s;        
}

